//! ve sistema (.ve-sistema:x86-64)
//!
//! Provides entry opcodes for System V calling convention, optimized for specific prototypes.
//!

// https://refspecs.linuxbase.org/elf/x86_64-abi-0.99.pdf

const std = @import("std");

/// Used for stack parameter passing.
pub const WordLimit = 128;
const AsmBufferLimit = 4096;
const ClassBufferLimit = 256;

const Class = enum {
    void, // Denotes empty types.
    integer,
    sse,
    sseup,
    x87,
    x87up,
    no_class,
    memory,
};

fn determiteClass(comptime T: type, buffer: []Class) []Class {
    switch (@typeInfo(T)) {
        .Void => &[1]Class{.void},
        .Int => |int| {
            switch (int.bits) {
                0 => buffer[0] = .void,
                1...64 => buffer[0] = .integer,
                65...128 => @compileError("unimplemented"),
                else => @compileError("unimplemented"),
            }
        },
        .Float => |float| {
            switch (float.bits) {
                0 => buffer[0] = .void,
                1...64 => buffer[0] = .sse,
                65...80 => @compileError("unimplemented"),
                81...128 => @compileError("unimplemented"),
                else => @compileError("unimplemented"),
            }
        },
        .Bool => buffer[0] = .integer,
        .Pointer => |ptr| {
            switch (ptr.size) {
                .Slice => {
                    buffer[0] = .integer;
                    buffer[1] = .integer;
                },
                else => buffer[0] = .integer,
            }
        },
        .Fn => buffer[0] = .integer,
        else => @compileError("unimplemented"),
    }

    return buffer[0 .. (@sizeOf(T) - 1) / 8 + 1];
}

// todo: Make sure duplicates are not made.
// todo: Cache results for identical in effect devices.
//
/// (iw | -- (arbitrary amount of words))
pub fn generateOpZovSysvFromPrototype(prototype: anytype) !*const fn () callconv(.Naked) noreturn {
    // todo: Should we care about this?
    // > The direction flag DF in the %rFLAGS register must be clear (set to “forward”
    // > direction) on function entry and return.

    comptime {
        const func = @typeInfo(@TypeOf(prototype)).Fn;

        var source_buffer = [_]u8{0} ** AsmBufferLimit;
        var source_needle: usize = 0;

        // todo: Align callee frame to 16?
        // > shrq     $4, %%rsp
        // > addq     $1, %%rsp
        // > shlq     $4, %%rsp

        // idea: Try using REP for big consequent memory pushes.

        // todo: In-stack returns by pointing %rdi directly to final destination.

        const Prelude =
            \\ movq %%rsp, %%rbp # Move stack pointer in non-volatile %rbp to restore later
            \\ subq $0x8, %%rsp
            \\
        ;

        const Call =
            \\ call *8(%%r12)
            \\
        ;

        const Epilogue =
            \\ movq %%rbp, %%rsp # Restore stack pointer
            \\ addq $0x10, %%r12
            \\ jmpq *(%%r12)
            \\
        ;

        @memcpy(source_buffer[source_needle .. source_needle + Prelude.len], Prelude[0..]);
        source_needle += Prelude.len;

        var integer_allocation: usize = 0;
        const IntegerAllocations = [_][]const u8{ "rdi", "rsi", "rdx", "rcx", "r8", "r9", "stack" };
        //  var sse_allocation: enum { xmm0, xmm1, xmm2, xmm3, xmm4, xmm5, xmm6, xmm7, stack } = .xmm0;

        var class_buffer = [_]Class{.void} ** ClassBufferLimit;

        // Calculate stack space used by parameters.
        var parameter_stack_size: usize = 0;
        for (func.params) |param| {
            const classes = determiteClass(param.type.?, &class_buffer);
            parameter_stack_size += 8 * classes.len;
        }

        // Push parameters to appropriate registers and stack positions.
        var stack_offset: usize = parameter_stack_size;
        for (func.params) |param| {
            const classes = determiteClass(param.type.?, &class_buffer);
            for (classes) |class| {
                stack_offset -= 8;
                switch (class) {
                    .integer => {
                        if (integer_allocation < IntegerAllocations.len - 1) {
                            source_needle += (try std.fmt.bufPrint(
                                source_buffer[source_needle..],
                                "movq {}(%%rbp), %%{s}\n",
                                .{ stack_offset, IntegerAllocations[integer_allocation] },
                            )).len;
                            integer_allocation += 1;
                        } else {
                            source_needle += (try std.fmt.bufPrint(
                                source_buffer[source_needle..],
                                "pushq {}(%%rbp)\n",
                                .{stack_offset},
                            )).len;
                        }
                    },
                    .void => {},
                    else => @compileError("unimplemented"),
                }
            }
        }

        @memcpy(source_buffer[source_needle .. source_needle + Call.len], Call[0..]);
        source_needle += Call.len;

        @memcpy(source_buffer[source_needle .. source_needle + Epilogue.len], Epilogue[0..]);
        source_needle += Epilogue.len;

        return &struct {
            fn op() callconv(.Naked) noreturn {
                asm volatile (source_buffer[0..source_needle]);
            }
        }.op;
    }
}
